# -*- coding:utf-8 -*-
# Copyright (C) 2013 Deutsches Zentrum fuer Luft- und Raumfahrt(DLR, German Aerospace Center) <www.dlr.de>
"""
Created on 23.04.2013
@author: saue_a0
"""
import itertools
import os
import pickle

import numpy as np
from patme.service.logger import log
from patme.service.stringutils import indent

from delismm.model.customsystemfunction import BoundsHandler
from delismm.model.samplecalculator import getY
from delismm.service.exception import DelisMMError


class DOEQuality:
    """Assesses a DOEs quality

    [1]M. D. Morris and T. J. Mitchell, "Exploratory designs for computational experiments," Journal of statistical planning and inference, vol. 43, no. 3, pp. 381–402, 1995.


    """

    @staticmethod
    def jd(sampleXNormalized):
        """computes the multiplicity array and the distinct distances of an LH#
        @param lhs the sampling plan
        """

        lhs = sampleXNormalized.T
        n, p = lhs.shape
        first = np.zeros((n * (n - 1) // 2, p))
        second = np.zeros((n * (n - 1) // 2, p))

        start = 0
        for i in range(n - 1):
            length = n - i - 1
            first[start : start + length, :] = np.zeros((length, p)) + lhs[i]
            second[start : start + length, :] = lhs[i + 1 :, :]
            start += length
        d = np.linalg.norm(first - second, axis=1)

        distinct, J = np.lib.arraysetops.unique(d, return_counts=True)
        return J, distinct

    @staticmethod
    def mm(sampleXNormalized):
        """Computes the maximin metric
        computes sampling plan quality according to Morris&Mitchell
        @param lhs the sampling plan"""
        J, d = DOEQuality.jd(sampleXNormalized)
        return np.sum(np.multiply(J, d**-2)) ** 0.5


class AbstractDOE:
    """Abstract class with interface specifications for DOE methods"""

    def __init__(self, n, p):
        """Initializes the DOE parameters n and p, where n is the number of sample points and p is the dimension of the design space"""
        self.n = n  # n is the desired number of points
        self.p = p  # p is the number of design variables, thus the dimension of design space
        self.sampleXNormalized = None
        """matrix with samples created with length [p, n]
        This matrix is always normalized [0,1]"""

    def _runDOE(self):
        """Method that creates a DOE with n sample points for the unit cube [0,1]^p"""
        raise NotImplementedError("This method must be implemented in a subclass")

    def getY(self, targetFunctionInstance, verbose=True):
        """see getY as description"""
        sampleX = BoundsHandler.scaleToBoundsStatic(
            self.sampleXNormalized, targetFunctionInstance.lowerBounds, targetFunctionInstance.upperBounds
        )
        return np.array(getY(sampleX, targetFunctionInstance, verbose))

    def xToFile(self, filename):
        """Method that saves sample points to a given path (but not the function values!)
        :param filename: filename of the file in which the sample points shall be saved
        """
        self.__class__.xToFileStatic(filename, self.sampleXNormalized)

    @staticmethod
    def xToFileStatic(filename, sampleXNormalized):
        """Method that saves sample points to a given path (but not the function values!)
        :param filename: filename of the file in which the sample points shall be saved
        """
        np.savetxt(filename, sampleXNormalized.T)

    @staticmethod
    def yToFile(filename, targetFunctionInstance, sampleY):
        """Saves the function values of the sample points to a given path (but not the sample points!).
        :param filename: filename of the file in which the function values shall be saved
        :param targetFunctionInstance: instance of a delismm.model.customsystemfunction class
            or a sub class of it.
        :param sampleY: sampleY that should be saved
        """
        header = ""
        if targetFunctionInstance is not None:
            header += str(targetFunctionInstance.__class__) + "\n"
            if targetFunctionInstance.bounds:
                header += "Parameters: \n" + indent(targetFunctionInstance.bounds)
                if hasattr(targetFunctionInstance, "resultNames"):
                    header += "Result values: \n{}".format(", ".join(targetFunctionInstance.resultNames)) + "\n"
        header = "# " + header.replace("\n", "\n# ") + "\n"
        body = indent(sampleY, delim=" ")
        with open(filename, "w") as f:
            f.write(header + body)
        pickleFilename = os.path.splitext(filename)[0] + ".pickle"
        with open(pickleFilename, "wb") as f:
            pickle.dump(sampleY, f)

    @staticmethod
    def ysFromFile(sampleYFile, targetFunctionInstance=None):
        """Reads the function values of the sample points from a given path (but not the sample points!).
        :param sampleYFile: filename of the file in which the function values shall be read
        :param targetFunctionInstance: instance of a delismm.model.customsystemfunction class
            or a subclass of it.
        :return: sampleY that should be read
        """
        if not os.path.exists(sampleYFile):
            log.info("Load sampleY failed. This path and file does not exist: " + sampleYFile)
            return None
        if os.path.exists(sampleYFile) and ".pickle" in sampleYFile:
            with open(sampleYFile, "rb") as f:
                sampleY = pickle.load(f)
            if targetFunctionInstance is not None and hasattr(targetFunctionInstance, "resultNames"):
                if sampleY[0] is None or len(sampleY[0]) != len(targetFunctionInstance.resultNames):
                    log.info(
                        f"sampleY does not contain same number of results as given target function. \
                                 SampleY[0]: {sampleY[0]}, target function result names: {targetFunctionInstance.resultNames}"
                    )
                    return []
            log.info("Read sampleY, count: " + str(len(sampleY)))
            return sampleY
        else:
            # old version
            ys = np.loadtxt(sampleYFile)
            if ys.ndim == 1:
                # old format
                ys = ys.reshape((len(ys), 1))
            return ys

    def xyToFile(self, filename, sampleY, headerNames=None, lb=None, ub=None, scaleToBounds=False):

        allSamples = [] if headerNames is None else [headerNames]
        allSamples.extend(self._getSamplesForXYToFile(sampleY, lb, ub, scaleToBounds))
        with open(filename, "w") as f:
            f.write(indent(allSamples, delim="  ", hasHeader=headerNames is not None))

    def _getSamplesForXYToFile(self, sampleY, lb=None, ub=None, scaleToBounds=False):
        """Saves the sample points to a given path.
        :param sampleY: if you already have evaluated your sample points, you don't have to evaluate them
            again. Just pass the function values as the parameter sampleY.
        :param lb: lower bounds, dict{key:lowerbound}
        :param ub: upper bounds, dict{key:upperbound}
        :param scaleToBounds: Flag if sampleXNormalized should be scaled to original bounds
        :return: list with all samples
        """
        allSamples = []
        sampleX = (
            self.sampleXNormalized
            if not scaleToBounds
            else BoundsHandler.scaleToBoundsStatic(self.sampleXNormalized, list(lb.values()), list(ub.values()))
        )
        for inputSample, outputSample in zip(sampleX.T, sampleY):
            if hasattr(outputSample, "__iter__"):
                allSamples.append(list(inputSample) + list(outputSample))
            else:
                allSamples.append(list(inputSample) + list([outputSample]))
        return allSamples

    @staticmethod
    def xyToFileForPgfPlotsStatic(samples, filename, numberOfSplits=None, wrapFunc=None):
        """Creates a file writing the samples in the format for pgfplot input files.

        :param samples: some matrix-like iterable. It is the matrix of samples in the transponed form as doe.sampleXNormalized or not normalized
        :param filename: filename where the new file will be created
        :param numberOfSplits: in case of a full factorial design, this is the number of splits in each dimension
        :param wrapFunc: function that can be used to modify each element in samples while it is put in the list
        """
        samples = list(samples)
        if numberOfSplits is not None:
            # if splits are given, a full factorial design is assumed, then a countour plot is possibly used, scatter plot otherwise
            for sampleIndex in range(len(samples) - numberOfSplits, 0, -1 * numberOfSplits):
                samples.insert(sampleIndex, ["", ""])
        kwargs = {"delim": "  "}
        if wrapFunc is not None:
            kwargs["wrapfunc"] = wrapFunc
        with open(filename, "w") as f:
            f.write(indent(samples, **kwargs))

    def __add__(self, otherDoeOrSampleXNormalized):
        """Adds another DOE or additional Samples with this DOE creating a new class."""
        if isinstance(otherDoeOrSampleXNormalized, AbstractDOE):
            p = otherDoeOrSampleXNormalized.p
            n = self.n + otherDoeOrSampleXNormalized.n
            additionalSamples = otherDoeOrSampleXNormalized.sampleXNormalized
        elif isinstance(otherDoeOrSampleXNormalized, np.ndarray):
            p = otherDoeOrSampleXNormalized.shape[0]
            n = self.n + otherDoeOrSampleXNormalized.shape[1]
            additionalSamples = otherDoeOrSampleXNormalized
        else:
            raise DelisMMError("The given parameter must be either a doe or a numpy array")
        if p != self.p:
            raise DelisMMError("The given doe has a different number of parameters that this doe.")
        newSampleXNormalized = np.zeros((self.p, n))
        newSampleXNormalized[:, : self.n] = self.sampleXNormalized
        newSampleXNormalized[:, self.n :] = additionalSamples
        newDoe = AbstractDOE(n, self.p)
        newDoe.sampleXNormalized = newSampleXNormalized
        return newDoe

    @staticmethod
    def latinizeSampleXNormalized(sampleXNormalized):
        """Latinizes the given samples. Each sample point must be set on a latin hypercube grid.

        See chapter 4 in
        [1] Y. Saka, M. Gunzburger, and J. Burkardt, "Latinized, improved LHS, and CVT point sets in hypercubes," International Journal of Numerical Analysis and Modeling, vol. 4, no. 3–4, pp. 729–743, 2007.
        """
        samples = sampleXNormalized.T
        n, p = samples.shape
        indicies = range(n)
        for k in range(p):
            sortVector = samples[:, k]
            indicies, samples = zip(
                *[(index, sample) for _, index, sample in sorted(zip(sortVector, indicies, samples))]
            )
            samples = np.array(samples)
            samples[:, k] = np.linspace(0, 1, n)

        # sort for initial index
        samples = np.array([sample for _, sample in sorted(zip(indicies, samples))])
        return samples.T

    @staticmethod
    def joinDoeResults(doeDirs, targetDoeDir, targetFunction=None, parameterNames=None):
        """combines sampleX and sampleY from inputDirs and puts them in targetDir"""
        xs = []
        ys = []
        for dirName in doeDirs:
            xs.extend(DOEfromFile(os.path.join(dirName, "sampleX_bounds.txt")).sampleXNormalized.T)
            ys.extend(DOEfromFile.ysFromFile(os.path.join(dirName, "sampleY.pickle")))

        sampleX, sampleY = [], []
        for x, y in zip(xs, ys):
            if y[0] is None:
                continue
            sampleX.append(x)
            sampleY.append(y)

        sampleX = np.array(sampleX).T
        doe = AbstractDOE(*sampleX.shape)
        doe.sampleXNormalized = sampleX  # not normalized
        doe.xToFile(os.path.join(targetDoeDir, "sampleX_bounds.txt"))
        doe.yToFile(os.path.join(targetDoeDir, "sampleY.txt"), targetFunction, sampleY)
        doe.xyToFile(
            os.path.join(targetDoeDir, "full_doe.txt"),
            sampleY,
            headerNames=([] if parameterNames is None else parameterNames)
            + ([] if targetFunction is None else targetFunction.resultNames),
        )


class DOEfromFile(AbstractDOE):
    def __init__(self, path):
        """Reads sample points from file
        :param path: path of the file
        """
        if not os.path.exists(path):
            raise DelisMMError("This path does not exist: " + path)
        self.path = path
        self.sampleXNormalized = self._runDOE()
        log.debug(
            "sampleXNormalized = \n" + indent(self.sampleXNormalized) + " type: " + str(type(self.sampleXNormalized))
        )
        self.n = self.sampleXNormalized.shape[1]
        self.p = self.sampleXNormalized.shape[0]

    def _runDOE(self):
        """See AbstractDOE for documentation"""
        newSampleXNormalized = np.loadtxt(self.path)
        if len(newSampleXNormalized.shape) == 1:
            # if p == 1 then numpy writes only sampleXNormalized[:,] to file
            return np.array([newSampleXNormalized])
        else:
            return newSampleXNormalized.T


class MonteCarlo(AbstractDOE):
    """Generates a random sample set [0,..,1]^p, where p is the number of design variables."""

    def __init__(self, n, p):
        AbstractDOE.__init__(self, n, p)
        self.sampleXNormalized = self._runDOE()

    def _runDOE(self):
        """See AbstractDOE for documentation"""
        return np.random.rand(self.p, self.n)


class LatinHypercube(AbstractDOE):
    """Generates a random Latin Hypercube within [0,..,1]^p, where p is the number of design variables."""

    def __init__(self, n, p):
        AbstractDOE.__init__(self, n, p)
        self.sampleXNormalized = self._runDOE()

    def _runDOE(self):
        """See AbstractDOE for documentation"""

        sampleXNormalized = np.zeros((self.p, self.n))  # Erzeuge zunaechst eine p-dim Liste mit Samplepoints
        for i in range(self.p):  # Fuer jede der Designvariablen erstelle
            x = np.arange(self.n)  # Vektor mit Zahlen von 0 bis n-1
            np.random.shuffle(x)  # shuffle des Vektors
            x = x / float(self.n - 1)
            sampleXNormalized[i, :] = x
        # In sampleXNormalized stehen nun Spaltenweise die Koordinaten der Samplepoints, i-th sample-coordinates: sampleXNormalized[:,i]
        return sampleXNormalized


class FullFactorialDesign(AbstractDOE):
    """Generates a full factorial design within [0,..,1]^p

    :param n: number of samples. This parameter might be adjusted if n!=i^p for i=1..inf
            If n does not satisfy the equation the next lower value is used
    """

    def __init__(self, n, p):
        if p == 0:
            n = 0
        numberOfSplits = 1
        while numberOfSplits**p < n + 1:
            numberOfSplits += 1
        numberOfSplits -= 1
        n = numberOfSplits**p
        self.numberOfSplits = numberOfSplits
        AbstractDOE.__init__(self, n, p)
        self.sampleXNormalized = self._runDOE()

    def _runDOE(self):
        sampleXNormalized = np.array(
            [it.flatten() for it in np.meshgrid(*([np.linspace(0.0, 1.0, self.numberOfSplits)] * self.p))]
        )
        # somehow meshgrid confuses first and second axes
        if self.p > 1:
            sampleXNormalized[:2, :] = np.array([sampleXNormalized[1, :], sampleXNormalized[0, :]])
        return sampleXNormalized

    def _runDOEOld(self):
        """doc"""
        sampleXNormalized = np.zeros((self.p, self.numberOfSplits**self.p))
        j = 0
        # product returns all combinations of Listobjects in [] with repeats and return length equal to 'repeat'
        for L in itertools.product(np.linspace(0.0, 1.0, self.numberOfSplits), repeat=self.p):
            sampleXNormalized[:, j] = L
            j += 1
        return sampleXNormalized

    @staticmethod
    def isFFDSamples(sampleX, lowerBounds, upperBounds):
        """Returns True if the given samples are based on a full factorial design"""
        sampleXNormalized = BoundsHandler.scaleTo01Static(sampleX, lowerBounds, upperBounds)
        numberOfParameters, numberOfSamples = sampleXNormalized.shape
        first = sampleXNormalized[:, 0]
        last = sampleXNormalized[:, -1]
        if not np.allclose(first, np.zeros_like(first)) or not np.allclose(last, np.ones_like(first)):
            return False
        numberOfSplits = 2
        while numberOfSplits**numberOfParameters < numberOfSamples:
            numberOfSplits += 1
        if numberOfSplits**numberOfParameters != numberOfSamples:
            return False
        return True


class ScreeningDoe(AbstractDOE):
    """Morris' algorithm(1991) described in Forrester: "Engineering Design via Surrogate Modelling" chapter 1.3.1.

    :param n: number of samples. This parameter might be adjusted if n != (p+1)*i for i=1..inf
            If n does not satisfy the equation the next lower value is used. In the book this parameter is r * (self.p+1) = self.n
    :param p: number of design parameters. In the book this parameter is "k"
    :param xi: elementary effect step length factor.
    :param l: number of discrete levels along each dimension. In the book this parameter is "p"


    """

    def __init__(self, n, p, xi=1, l=21):
        n = n - divmod(n, p + 1)[1]
        self.xi = xi
        self.l = l
        AbstractDOE.__init__(self, n, p)
        self.sampleXNormalized = self._runDOE()

    def _runDOE(self):
        r = self.n // (self.p + 1)
        sampleXNormalized = self._screeningPlan(self.p, self.l, self.xi, r)
        return sampleXNormalized

    def _randorient(self, p, l, xi):
        """Generates a random orientation for a screening matrix

        :param p: number of design variables
        :param l: number of discrete levels along each dimension
        :param xi: elementary effect step length factor

        :return: Bstar - random orientation matrix
        """

        # Step length
        delta = xi / (l - 1.0)

        m = p + 1

        # A truncated p-level grid in one dimension
        xs = np.arange(0.0, 1 - delta + 1.0 / (l - 1), 1.0 / (l - 1))  # (0:1/(p-1):1-delta)
        xsl = len(xs)

        # Basic sampling matrix
        b = np.tril(np.ones((m, p)), -1)

        # Randomization
        # Matrix with +1s and -1s on the diagonal with equal probability
        # in Diss C*
        dStar = np.diag(np.round(np.random.uniform(size=p)) * 2 - 1)

        # Random base value
        xStar = (np.floor(np.random.uniform(size=p) * xsl)) / xsl

        # Permutation matrix; in Diss this is A*
        pStar = np.zeros((p, p))
        rp = np.random.permutation(p)
        for index in range(p):
            pStar[index, rp[index]] = 1.0

        # A random orientation of the sampling matrix
        bStar = np.dot(
            (
                np.dot(np.ones((m, 1)), np.array([xStar]))
                + (delta / 2.0) * (np.dot((2 * b - np.ones((m, p))), dStar) + np.ones((m, p)))
            ),
            pStar,
        )
        # (ones(m,1)*xstar+(Delta/2)*((2*B-ones(m,k))*Dstar+ones(m,k)))*Pstar;
        return bStar

    def _screeningPlan(self, p, l, xi, r):
        """Generates a Morris screening plan with a specified number of elementary effects for each variable.

        :param p: number of design variables
        :param l: number of discreet levels along each dimension
        :param xi: elementary effect step length factor
        :param r: number of random orientations (=number of elementary effects per variable).

        :return: screening plan with values of [0,1] of the shape (k+1*r, k)
        """
        retArray = self._randorient(p, l, xi)
        for index in range(r - 1):
            retArray = np.vstack((retArray, self._randorient(p, l, xi)))

        return retArray.T


class CentroidalVoronoiTesselation(AbstractDOE):
    """
    [1]Y. Saka, M. Gunzburger, and J. Burkardt, "Latinized, improved LHS, and CVT point sets in hypercubes," International Journal of Numerical Analysis and Modeling, vol. 4, no. 3–4, pp. 729–743, 2007.
    [2]L. Ju, Q. Du, and M. Gunzburger, "Probabilistic methods for centroidal Voronoi tessellations and their parallel implementations," Parallel Computing, vol. 28, no. 10, pp. 1477–1500, 2002.
    """

    doPlot = False

    def __init__(self, n, p, cellEstimatorCountFactor=5, maxIteration=1000, requiredStableRuns=10, initialDoe=None):
        """Also See AbstractDOE for documentation

        :param cellEstimatorCountFactor: The number of cell estimators can be calculated with this: n * cellEstimatorCountFactor.
            This value may be a float but greater zero.
        """
        self.doeGeneratorClass = MonteCarlo
        AbstractDOE.__init__(self, n, p)
        if cellEstimatorCountFactor <= 0:
            raise DelisMMError("cellEstimatorCountFactor is zero or negative")
        self.sampleXNormalized = self._runDOE(cellEstimatorCountFactor, maxIteration, requiredStableRuns, initialDoe)

    def _runDOE(self, cellEstimatorCountFactor, maxIteration, requiredStableRuns, initialDoe):
        """
        variables in [2]:

            z = generators
            q = cellEstimators
        """

        from scipy.spatial import cKDTree

        nCellEstimators = self.n * cellEstimatorCountFactor
        alpha, beta = (0.95, 0.05), (0.5, 0.5)
        if not alpha[1] >= 0 or not beta[1] >= 0 or not np.allclose([sum(alpha), sum(beta)], [1, 1]):
            raise DelisMMError("values of alpha and beta are malformed")

        # initial population
        if initialDoe is None:
            newGenerators = self.doeGeneratorClass(self.n, self.p).sampleXNormalized.T
        else:
            if initialDoe.n != self.n or initialDoe.p != self.p:
                raise DelisMMError("wrong doe parameters for initial doe")
            newGenerators = initialDoe.sampleXNormalized.T
        doeQuality0 = DOEQuality.mm(newGenerators.T)
        stable = 0
        best = None
        if self.doPlot:
            generatorsHistory = [newGenerators]
        log.info("cvt iterations start")
        cvtIteration = 0
        doeQuality = DOEQuality.mm(newGenerators.T)
        log.debug("CVT generation {}, quality of {}".format(cvtIteration, doeQuality))
        while 1:
            generators = newGenerators
            j = cvtIteration  # + 1
            c1, c2 = (alpha[0] * j + beta[0]) / (j + 1), (alpha[1] * j + beta[1]) / (j + 1)
            # create cell estimators
            cellEstimators = self.doeGeneratorClass(nCellEstimators, self.p).sampleXNormalized.T
            #             cellEstimators = MonteCarlo(nCellEstimators, self.p).sampleXNormalized.T

            # find nearest generators to each cell estimator
            tree = cKDTree(generators)
            _, nearestIndexes = tree.query(cellEstimators)

            # collect all cell estimators to each generator
            cellEstimatorDict = {}
            for nearestIndex, cellEstimator in zip(nearestIndexes, cellEstimators):
                if nearestIndex not in cellEstimatorDict:
                    cellEstimatorDict[nearestIndex] = []
                cellEstimatorDict[nearestIndex].append(cellEstimator)

            # calculate center(mean) of cell estimators for each voronoi cell
            cellCenters = generators.copy()

            for generatorIndex in cellEstimatorDict:
                cellCenter = np.mean(cellEstimatorDict[generatorIndex], axis=0)
                cellCenters[generatorIndex, :] = cellCenter
            newGenerators = c1 * generators + c2 * cellCenters
            cvtIteration += 1

            if self.doPlot:
                self._plotIteration(
                    generators, cellEstimators, nearestIndexes, newGenerators, cellCenters, cvtIteration
                )
                generatorsHistory.append(newGenerators)

            # ===================================================================
            # check quality and possibly converge
            # ===================================================================
            doeQuality = DOEQuality.mm(newGenerators.T)
            log.debug("CVT generation {}, quality of {}".format(cvtIteration, doeQuality))
            if doeQuality < doeQuality0:
                stable = 0
                best = generators
                doeQuality0 = doeQuality
            else:
                stable += 1
            if stable > requiredStableRuns and best is not None:
                log.info("Stable run at CVT generation {}, quality of {}".format(cvtIteration, doeQuality))
                break
            if cvtIteration >= maxIteration:
                log.info("MaxIteration reached with quality of {}".format(doeQuality))
                best = generators
                break

        doeQuality = DOEQuality.mm(newGenerators.T)

        if 1 and self.doPlot:
            self._plotIteration(generators, cellEstimators, nearestIndexes, newGenerators, cellCenters, cvtIteration)
            self._plotGeneratorTrace(generatorsHistory)
            import matplotlib.pyplot as plt

            plt.show()
        return best.T

    @staticmethod
    def _plotIteration(
        generators, cellEstimators, nearestIndexes=None, newGenerators=None, cellCenters=None, iteration=None
    ):
        """doc"""

        if generators.shape[1] != 2:
            log.info("not of dimension p=2")
            return
        import matplotlib.pyplot as plt

        fig = plt.figure(figsize=(8, 6))
        ax = plt.gca()

        plt.scatter(cellEstimators.T[0], cellEstimators.T[1], s=20, c="b", alpha=0.5, label="cell estimators")
        plt.scatter(generators.T[0], generators.T[1], s=100, c="r", alpha=0.5, label="old generators")
        if newGenerators is not None:
            plt.scatter(newGenerators.T[0], newGenerators.T[1], s=100, c="g", alpha=0.5, label="new generators")
            for generator, newGenerator in zip(generators, newGenerators):
                ax.annotate(
                    "",
                    xy=newGenerator,
                    xycoords="data",
                    xytext=generator,
                    textcoords="data",
                    weight="bold",
                    color="g",
                    arrowprops=dict(arrowstyle="->", connectionstyle="arc3", color="g"),
                )
        if cellCenters is not None:
            plt.scatter(cellCenters.T[0], cellCenters.T[1], s=100, c="b", alpha=0.5, label="cell center")
            if nearestIndexes is not None:
                for estimator, generatorIndex in zip(cellEstimators, nearestIndexes):
                    cellCenter = cellCenters[generatorIndex]
                    plt.plot(*zip(estimator, cellCenter), c="black")
            if newGenerators is not None:
                for cellCenter, newGenerator in zip(cellCenters, newGenerators):
                    plt.plot(*zip(newGenerator, cellCenter), "g--")
        CentroidalVoronoiTesselation._setAxisLimits(ax)

        # shrink plot region by 20 percent
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        # Put a legend to the right of the current axis
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))

        print("generators")
        print(indent(generators, delim=" , "))
        print("newGenerators")
        print(indent(newGenerators, delim=" , "))
        print("cellEstimators")
        print(indent(cellEstimators, delim=" , "))
        print("cellCenters")
        print(indent(cellCenters, delim=" , "))
        print("nearestIndexes")
        print(indent([nearestIndexes], delim=" , "))

        fig.savefig("cvt_iteration_{}.png".format(iteration))

    def _plotGeneratorTrace(self, generatorsHistory):
        """doc"""
        if generatorsHistory and generatorsHistory[0].shape[1] != 2:
            log.info("not of dimension p=2")
            return
        import matplotlib.pyplot as plt

        plt.figure()

        lastGenerators = generatorsHistory[0]
        for generators, transparency in zip(generatorsHistory[1:], np.linspace(0, 0.7, len(generatorsHistory) - 1)):
            for generator, lastGenerator in zip(lastGenerators, generators):
                plt.plot(*zip(generator, lastGenerator), c="black", alpha=transparency)
            # plt.plot(*zip(lastGenerators, generators), c = 'black', alpha=transparency)
            # plt.scatter(generators.T[0], generators.T[1], s=10, c='r', alpha=transparency)
            lastGenerators = generators
        plt.scatter(generatorsHistory[-1].T[0], generatorsHistory[-1].T[1], s=100, c="r", alpha=1)
        self._setAxisLimits(plt.gca())

    @staticmethod
    def _setAxisLimits(ax):
        """doc"""
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)


class LatinizedCentroidalVoronoiTesselation(CentroidalVoronoiTesselation):
    """Latinized version of CVT

    For latinization see chapter 4 in
    [1] Y. Saka, M. Gunzburger, and J. Burkardt, "Latinized, improved LHS, and CVT point sets in hypercubes," International Journal of Numerical Analysis and Modeling, vol. 4, no. 3–4, pp. 729–743, 2007.
    """

    def __init__(self, *args, **kwargs):
        """Also See AbstractDOE for documentation"""
        CentroidalVoronoiTesselation.__init__(self, *args, **kwargs)
        samples = self.sampleXNormalized
        self.sampleXNormalized = self.latinizeSampleXNormalized(self.sampleXNormalized)
        doeQuality = DOEQuality.mm(self.sampleXNormalized)
        log.info("LCVT quality of {}".format(doeQuality))

        if self.doPlot:
            self._plotLatinization(samples, self.sampleXNormalized)

    def _plotLatinization(self, sampleXBefore, sampleXAfter):
        """doc"""
        if sampleXBefore.shape[0] != 2:
            log.info("not of dimension p=2")
            return

        import matplotlib.pyplot as plt

        plt.figure()
        plt.scatter(sampleXBefore.T[:, 0], sampleXBefore.T[:, 1], s=100, c="r", alpha=1)
        plt.scatter(sampleXAfter.T[:, 0], sampleXAfter.T[:, 1], s=100, c="b", alpha=1)
        self._setAxisLimits(plt.gca())
        plt.show()


class ScaledCentroidalVoronoiTesselation(CentroidalVoronoiTesselation):
    """The CVT values are inside the unit cube. The samples are scaled to [0,1]"""

    def __init__(self, *args, **kwargs):
        """Also See AbstractDOE for documentation"""
        CentroidalVoronoiTesselation.__init__(self, *args, **kwargs)
        self.scaleSamples()

    def scaleSamples(self):
        """doc"""

        newSamples = []
        samples = self.sampleXNormalized
        for paramSamples in samples:
            minVal, maxVal = np.min(paramSamples), np.max(paramSamples)
            paramSamples = (paramSamples - minVal) / (maxVal - minVal)
            newSamples.append(paramSamples)

        newSamples = np.array(newSamples)
        self.sampleXNormalized = newSamples

        doeQuality = DOEQuality.mm(self.sampleXNormalized)
        log.info("SCVT quality of {}".format(doeQuality))

        if self.doPlot:
            self._plotLatinization(samples, self.sampleXNormalized)

    def _plotSampleSets(self, sampleXBefore, sampleXAfter):
        """doc"""
        if sampleXBefore.shape[0] != 2:
            log.info("not of dimension p=2")
            return

        import matplotlib.pyplot as plt

        plt.figure()
        plt.scatter(sampleXBefore.T[:, 0], sampleXBefore.T[:, 1], s=100, c="r", alpha=1)
        plt.scatter(sampleXAfter.T[:, 0], sampleXAfter.T[:, 1], s=100, c="b", alpha=1)
        self._setAxisLimits(plt.gca())
        plt.show()


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    if 0:
        doe = ScaledCentroidalVoronoiTesselation(20, 2)
        initialSamples = doe.sampleXNormalized
        doe.scaleSamples()
        doe._plotSampleSets(initialSamples, doe.sampleXNormalized)
    elif 0:
        doe = LatinHypercube(5, 2)
        print(indent(doe.sampleXNormalized, delim=", "))
    elif 1:
        log.logLevel = 10
        CentroidalVoronoiTesselation.doPlot = True
        doe = CentroidalVoronoiTesselation(
            9, 2, requiredStableRuns=100
        )  # , maxIteration = 300, cellEstimatorCountFactor = 10)

        samples = np.array([np.linspace(0, 1, 100), np.linspace(0, 1, 100)])
        print(samples)
        print(DOEQuality.mm(samples))
    elif 1:
        generators = np.array(
            [
                [0.76629408533, 0.166305226647],
                [0.293062940846, 0.681128562962],
                [0.726354951027, 0.788534292217],
                [0.402460538682, 0.0923247590001],
            ]
        )
        newGenerators = [
            [0.756860179899, 0.239727887205],
            [0.281123104862, 0.732778252933],
            [0.726354951027, 0.788534292217],
            [0.378261243462, 0.134339794939],
        ]
        cellEstimators = [
            [0.492972608182, 0.374410925403],
            [0.397927746531, 0.274669721494],
            [0.731988974671, 0.433296719586],
            [0.065332619838, 0.207614376813],
            [0.513328621031, 0.104124972234],
            [0.390262761025, 0.864705490092],
            [0.109027949879, 0.873185744714],
            [0.102753912002, 0.264713543399],
        ]
        cellCenters = [
            [0.731988974671, 0.433296719586],
            [0.249645355452, 0.868945617403],
            [0.726354951027, 0.788534292217],
            [0.314463101517, 0.245106707869],
        ]
        nearestIndexes = [3, 3, 0, 3, 3, 1, 1, 3]

        CentroidalVoronoiTesselation._plotIteration(
            np.array(generators),
            np.array(cellEstimators),
            np.array(nearestIndexes),
            np.array(newGenerators),
            np.array(cellCenters),
        )
        plt.show()



"""
Created on 17.02.2016

@author: freu_se
"""

import itertools
from datetime import datetime

from delismm.service.exception import DelisMMError


def getTimeString(useMilliSeconds=False):
    """returns a time string of the format: yyyymmdd_hhmmss"""
    dt = datetime.now()
    return dt.strftime("%Y%m%d_%H%M%S") + ("_{}".format(dt.microsecond) if useMilliSeconds else "")


def exportSampleDataForOptimus(sampleX, sampleY, xKeys, yKeys, filePath=None, outputType="training"):
    """Converts sample data into input test or training data for the program Optimus.
    :param sampleX: is an array in which contains the sample points column by column (one parameter configuration = one column)
    :param sampleY: is an array in which contains the scalar function values at the sample points
    :param xKeys: list of sampleX parameters (len(list) = number of columns of sampleX)
    :param yKeys: list of sampleY parameters (len(list) = len(sampleY))
    :param filePath: file path for output file (outputType will be added)
    :param outputType: switch to choose between training or test data output
    """
    if not filePath:
        filePath = "optimusInput.dat"
    filePath = filePath[: filePath.rfind(".")] + "_" + outputType + filePath[filePath.rfind(".") :]

    if outputType == "training":
        output = str(len(xKeys)) + "\n" + str(len(yKeys)) + "\n"
        output += " ".join([str(key) for key in xKeys + yKeys]) + "\n"
    elif outputType == "test":
        output = ""
    else:
        raise DelisMMError("outputType: " + outputType + " not supported. Use [training, test].")

    for xData, yData in zip(sampleX.T, sampleY):
        output += " ".join([str(x) for x in xData]) + " "
        output += str(yData) + "\n"
    f = open(filePath, "w")
    f.write(output)
    f.close()


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=1)



import os
from threading import Lock

import numpy as np
import requests
from bokeh.io import curdoc
from bokeh.layouts import column, row
from bokeh.models import ColorBar, ColumnDataSource, Div, HoverTool, LinearColorMapper, Select, Slider
from bokeh.plotting import figure, show
from bokeh.transform import transform
from patme.service.systemutils import log

from delismm import name
from delismm.model.apimodels import SurrogateInfo
from delismm.model.surrogate import Kriging
from delismm.service.exception import DelisMMError

# import logging
# log.logLevel = logging.DEBUG

# check environment variable
if os.environ.get("DELISMMSERVER") is not None:
    server = os.environ["DELISMMSERVER"]
else:
    server = f"https://delismm.nimbus.dlr.de"
actualSurrogate = None
callEndpoint = "/call/"
surrogateNames = list(requests.get(f"{server}/list_surrogates").json().values())
replotLock = False  # lock if anything should happen, when changing the active parameter
bokehTools = "pan,wheel_zoom,box_zoom,reset,save,undo"
plotPoints = ["6", "11", "21", "51", "101", "201", "501", "1001"]


def resetSurrogate():
    """Get the information of a surrogate model from the server."""
    surrogateName = getSurrogateSelector().value
    response = requests.get(f"{server}/surrogate_info/{surrogateName}")
    surrogateInfo = SurrogateInfo(**response.json())
    global actualSurrogate
    actualSurrogate = Kriging(
        np.zeros((surrogateInfo.number_of_parameters, 1)),
        [0],
        np.array(surrogateInfo.lowerBounds),
        np.array(surrogateInfo.upperBounds),
        parameterNames=surrogateInfo.parameterNames,
        resultNames=surrogateInfo.model_name,
    )


def getActiveSliderValues():
    sliders = getSliders()
    numberOfParameters = actualSurrogate.p
    return [slider.value for slider in sliders[:numberOfParameters]]


def updateSurrogateValues1D(variableParameterName1):
    global actualSurrogate
    log.debug("get 1Dplot data")
    surrogateName = getSurrogateSelector().value
    numberOfPlotPoints = getNumberOfPlotPoints()
    allParameters = np.array([getActiveSliderValues()] * numberOfPlotPoints)
    paramIndex1 = actualSurrogate.parameterNames.index(variableParameterName1)
    x = np.linspace(
        actualSurrogate.lowerBounds[paramIndex1], actualSurrogate.upperBounds[paramIndex1], numberOfPlotPoints
    )
    allParameters[:, paramIndex1] = x
    results = []
    for parameters in allParameters:
        request = (
            f"{server}{callEndpoint}?surrogate_name={surrogateName}&"
            f"parameters={'&parameters='.join([str(p) for p in parameters])}"
        )
        r = requests.get(request)
        results.append(float(r.text))
    source1D.data = dict(x=x, data=np.array(results))
    log.debug("done getting 1Dplot data")


def updateSurrogateValues2D(variableParameterName1, variableParameterName2):
    global actualSurrogate
    log.debug("get 2Dplot data")
    surrogateName = getSurrogateSelector().value
    numberOfPlotPoints = getNumberOfPlotPoints()
    allParameters = np.array([getActiveSliderValues()] * numberOfPlotPoints**2)
    paramIndex1 = actualSurrogate.parameterNames.index(variableParameterName1)
    paramIndex2 = actualSurrogate.parameterNames.index(variableParameterName2)
    x = np.linspace(
        actualSurrogate.lowerBounds[paramIndex1], actualSurrogate.upperBounds[paramIndex1], numberOfPlotPoints
    )
    y = np.linspace(
        actualSurrogate.lowerBounds[paramIndex2], actualSurrogate.upperBounds[paramIndex2], numberOfPlotPoints
    )
    gridx, gridy = np.meshgrid(x, y)
    gridx, gridy = gridx.flatten(), gridy.flatten()
    allParameters[:, paramIndex1] = gridx
    allParameters[:, paramIndex2] = gridy
    results = []
    for parameters in allParameters:
        request = (
            f"{server}{callEndpoint}?surrogate_name={surrogateName}&"
            f"parameters={'&parameters='.join([str(p) for p in parameters])}"
        )
        r = requests.get(request)
        results.append(float(r.text))
    source2D.data = dict(x=gridx, y=gridy, data=np.array(results))
    log.debug("done getting 2Dplot data")
    return x, y


def replot(createNewPlot=False):
    surrogateName = getSurrogateSelector().value
    parameterName1, parameterName2 = [paramWidget.value for paramWidget in getActiveParameters()]
    if createNewPlot:
        plot = figure(width=600, height=500, margin=30, toolbar_location="below", tools=bokehTools)
        curdoc().roots[0].children[1] = plot
    else:
        plot = getPlot()
    if parameterName2 == "none":
        updateSurrogateValues1D(parameterName1)
        if createNewPlot:
            plot.line("x", "data", source=source1D, line_width=3, line_alpha=0.6)
        plot.add_tools(HoverTool(tooltips=[(parameterName1, "@x"), (surrogateName, "@data")]))
        plot.yaxis.axis_label = surrogateName
    else:
        x, y = updateSurrogateValues2D(parameterName1, parameterName2)
        if createNewPlot:
            mapper = LinearColorMapper(
                palette="Viridis256", low=source1D.data["data"].min(), high=source1D.data["data"].max()
            )
            plot.rect(
                "x",
                "y",
                width=x[1] - x[0],
                height=y[1] - y[0],
                source=source2D,
                line_color=None,
                fill_color=transform("data", mapper),
            )
            plot.add_layout(ColorBar(color_mapper=mapper), "right")
        plot.add_tools(HoverTool(tooltips=[(parameterName1, "@x"), (parameterName2, "@y"), (surrogateName, "@data")]))
        plot.yaxis.axis_label = parameterName2

    plot.title.text = surrogateName
    plot.xaxis.axis_label = parameterName1
    log.debug(f"replot done")


def eventSurrogateChanged(attr, old, new):
    log.info(f"eventSurrogateChanged: Selected surrogate: {new}")
    if new not in surrogateNames:
        raise Exception(f"{new} is not in valid surrogate names: {surrogateNames}")
    resetSurrogate()
    numberOfParameters = actualSurrogate.p
    global replotLock
    for slider, paramName, lb, ub in zip(
        getSliders()[:numberOfParameters],
        actualSurrogate.parameterNames,
        actualSurrogate.lowerBounds,
        actualSurrogate.upperBounds,
    ):
        slider.visible = True
        slider.title = paramName
        slider.start = lb
        slider.end = ub
        slider.step = (slider.end - slider.start) / 1000
        slider.value = lb + (slider.end - slider.start) / 2
    for slider in getSliders()[numberOfParameters:]:
        slider.visible = False
    selectActiveParameter1, selectActiveParameter2 = getActiveParameters()
    selectActiveParameter1.options = actualSurrogate.parameterNames
    selectActiveParameter2.options = ["none"] + actualSurrogate.parameterNames
    replotLock = True
    setActiveParameter1(actualSurrogate.parameterNames[0])
    setActiveParameter2("none")
    replotLock = False
    replot()


def onchangeNumberOfSamples(attr, old, new):
    log.info(f"onchangeNumberOfSamples: {attr} changed from {old} to {new}")
    replot()


def onchangeActiveParameter1(attr, old, new):
    log.debug(f"onchangeActiveParameter1: {attr} changed from {old} to {new}. Lock is active: {replotLock}")
    if replotLock:
        return  # if lock is active, do nothing
    replot()


def onchangeActiveParameter2(attr, old, new):
    log.debug(f"onchangeActiveParameter2: {attr} changed from {old} to {new}. Lock is active: {replotLock}")
    if replotLock:
        return  # if lock is active, do nothing
    createNewPlot = False
    if old == "none":
        log.debug(f"switch plot from 1D to 2D")
        createNewPlot = True
    if new == "none":
        log.debug(f"switch plot from 2D to 1D")
        createNewPlot = True
    replot(createNewPlot=createNewPlot)


def onchangeSlider(attr, old, new):
    log.info(f"onchangeSlider: {attr} changed from {old} to {new}")
    replot()


def createWidgets():
    selectSurrogateName = Select(title="Select surrogate", options=surrogateNames, width=300)
    selectSurrogateName.on_change("value", eventSurrogateChanged)

    selectNumberOfSamples = Select(title="Number of Samples", options=plotPoints, value="6", width=300)
    selectNumberOfSamples.on_change("value", onchangeNumberOfSamples)

    selectActiveParameter = Select(options=[], width=300)
    selectActiveParameter.on_change("value", onchangeActiveParameter1)
    selectActiveParameter2 = Select(value="none", options=[], width=300)
    selectActiveParameter2.on_change("value", onchangeActiveParameter2)

    p = figure(width=600, height=500, margin=30, toolbar_location="below", tools=bokehTools)

    sliders = []
    for i in range(10):
        slider = Slider(start=0.1, end=10, value=1, step=0.1, title=f"Parameter {i}")
        slider.on_change("value_throttled", onchangeSlider)
        sliders.append(slider)

    layout = row(
        column(selectSurrogateName, selectNumberOfSamples),
        p,
        column(
            [
                Div(text="<b>Plot Parameters</b>"),
                selectActiveParameter,
                selectActiveParameter2,
                Div(text="\n<b>Nominal parameter values:</b>"),
            ]
            + sliders
        ),
    )
    return layout


def getSurrogateSelector():
    return curdoc().roots[0].children[0].children[0]


def getPlot():
    return curdoc().roots[0].children[1]


def getSliders():
    return curdoc().roots[0].children[2].children[4:]


def setNumberOfPlotPoints(value):
    if str(value) not in plotPoints:
        raise DelisMMError("Invalid number of plot points")
    curdoc().roots[0].children[0].children[1].value = value


def getNumberOfPlotPoints():
    return int(curdoc().roots[0].children[0].children[1].value)


def getActiveParameters():
    return curdoc().roots[0].children[2].children[1], curdoc().roots[0].children[2].children[2]


def setActiveParameter1(param):
    curdoc().roots[0].children[2].children[1].value = param


def setActiveParameter2(param):
    curdoc().roots[0].children[2].children[2].value = param


def setActiveParameters(param1, param2):
    global replotLock
    replotLock = True
    curdoc().roots[0].children[2].children[1].value = param1
    replotLock = False
    curdoc().roots[0].children[2].children[2].value = param2


source1D = ColumnDataSource(data=dict(x=[0, 1], y=[0, 2]))
source2D = ColumnDataSource()
layout = createWidgets()
curdoc().add_root(layout)
curdoc().title = name + " plotter"
dummyPlot = getPlot()
dummyPlot.line("x", "data", source=source1D, line_width=3, line_alpha=0.6)

if 1:
    getSurrogateSelector().value = "lh2_tank_totalMass[kg]_v_1.1"
    getSliders()[0].value = 1716 * 2
    getSliders()[1].value = 0.66
    getSliders()[2].value = 0.2
    # setActiveParameters("dcyl[mm]", "lcylByR-")

if __name__ == "__main__":
    show(curdoc().roots[0])



# -*- coding: utf-8 -*-
"""
api functions for delismm
"""

import errno
import os
from argparse import ArgumentParser
from collections import OrderedDict
from typing import List

import numpy as np
import uvicorn
from fastapi import FastAPI, HTTPException, Query
from patme import epsilon

from delismm import description, name, programDir, version
from delismm.model.surrogate import AbstractSurrogate

modelDir = os.path.join(programDir, "models")
loadedSurrogates = {}

# instantiate the api
app = FastAPI(title=name, description=description, version=version, docs_url="/documentation")


@app.get("/")
def read_root() -> dict:
    """
    This function points to the main root of the api

    Returns
    -------
    dict
        Hello World
    """
    return {"Hello": "World"}


@app.get("/list_surrogates")
def list_surrogates() -> dict:
    """
    Lists all surrogates present in the surrogate folder

    Returns
    -------
    dict
        Maps an incemented index to the surrogate name
    """

    surrogateNames = getSurrogateNames()
    surrogates = [(surrogateIndex, surrogateName) for surrogateIndex, surrogateName in enumerate(surrogateNames)]
    surrogateDict = OrderedDict(surrogates)
    return surrogateDict


@app.get("/surrogate_info/{surrogate_name}")
def surrogate_info(surrogate_name: str) -> dict:
    """
    Lists all surrogates present in the surrogate folder

    Parameters
    ----------
    - surrogate_name : string, name of the surrogate as returned from "list_surrogates"

    Raises
    ------
    HTTPException if the given surrogate_name does not exist

    Returns
    -------
    :return: dict
        Maps surrogate properites to their values
    """
    surrogate = getSurrogate(surrogate_name)
    info = surrogate.getKrigingDocumentationDict()
    return info.model_dump()


@app.get("/call/")
def call(surrogate_name: str = Query(None), parameters: List[float] = Query(None), boundsError: bool = Query(False)):
    """
    Calls a surrogate with its parameters

    Parameters
    ----------
    - surrogate_name : list with model parameters
    - parameters : list of parameters of the surrogate model
    - boundsError : bool, optional. Raises HTTPException if a parameter is out of bounds

    Raises
    ------
    HTTPException when
    - the given surrogate_name does not exist
    - when the number of given parameters does not match the surrogates number of parameters
    - the parameters are out of bounds and boundsError == True

    Returns
    -------
    dict
        surrogate result
    """
    surrogate = getSurrogate(surrogate_name)
    if len(parameters) != surrogate.p:
        raise HTTPException(
            status_code=404,
            detail=f"number of parameters do not match! "
            f"Surrogate uses {surrogate.p} parameters but got {len(parameters)}",
        )
    if boundsError:
        normalizedParameters = surrogate.getParametersNormalized(parameters)
        paramOutOfBounds = (normalizedParameters < -epsilon) | (normalizedParameters > 1 + epsilon)
        if any(paramOutOfBounds):
            boundsInfoList = [
                f"{lb} <= {p} <= {ub}" for lb, ub, p in zip(surrogate.lowerBounds, surrogate.upperBounds, parameters)
            ]
            raise HTTPException(
                status_code=404,
                detail=f"Parameter bounds error: {list(zip(surrogate.parameterNames, paramOutOfBounds, boundsInfoList))}",
            )
    res = surrogate(parameters)[0]
    return res


def getSurrogateNames():
    """returns a list of valid surrogate names"""
    return sorted(
        os.path.splitext(filename)[0] for filename in os.listdir(modelDir) if os.path.splitext(filename)[1] == ".pickle"
    )


def getSurrogate(surrogate_name):
    """returns the surrogate with the given name"""
    if surrogate_name in loadedSurrogates:
        return loadedSurrogates[surrogate_name]
    if surrogate_name not in getSurrogateNames():
        raise HTTPException(status_code=404, detail=f'no such surrogate: "{surrogate_name}"')
    surrogate = AbstractSurrogate.load(os.path.join(modelDir, surrogate_name + ".pickle"))
    loadedSurrogates[surrogate_name] = surrogate
    return surrogate


if __name__ == "__main__":
    ap = ArgumentParser(description="Starts app to calculate tank metamodel results")
    ap.print_usage = ap.print_help  # redirecting the print_usage method to the extended print_help method
    ap.add_argument("-d", "--dir", dest="dir", help="Relative Path to the actual directory", metavar="DIR")
    ap.add_argument("-p", "--port", dest="port", type=int, default=8000, help="Port of the web server")
    options = ap.parse_args()
    if options.dir:
        modelDir = os.path.join(programDir, options.dir)
        if not os.path.exists(modelDir):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), modelDir)
    uvicorn.run(app, host="0.0.0.0", port=options.port)



""""""

import os
from argparse import ArgumentParser

import numpy as np
from patme.service.logger import log

from delismm.model.doe import AbstractDOE
from delismm.model.metrics import relativeRootMeanSquareError, rootMeanSquareError
from delismm.model.surrogate import Kriging


class KrigingsCaller(list):
    def __call__(self, params):
        if len(params) != 3:
            raise Exception(f"number of given parameters must be 3 but got {len(params)}")
        res = []
        for mm in self:
            res.append(mm(params))
            params = params[:2]
        res = np.power(10, res)
        return res


def plotMetamodel(plotter, mm, lb, ub):
    plotSampeles = 150
    x = np.linspace(lb[0], ub[0], plotSampeles)
    y = np.linspace(lb[1], ub[1], plotSampeles)
    xv, yv = np.meshgrid(x, y)
    xv, yv = xv.flatten(), yv.flatten()
    z = np.array([mm([xVal, vVal]) for xVal, vVal in zip(xv, yv)])
    plotter.plotContour(xv, yv, z)


def writeData(outputDir, sampleX, sampleY):
    """write sampleX, sampleY"""
    sampleXFilename = os.path.join(outputDir, "sampleX_bounds.txt")
    if os.path.exists(sampleXFilename):
        log.info(f"Not writing sampleX file: already exists: {sampleXFilename}")
    else:
        np.savetxt(sampleXFilename, sampleX)
    sampleYFilename = os.path.join(outputDir, "sampleY.txt")
    if os.path.exists(sampleYFilename):
        log.info(f"Not writing sampleY file: already exists: {sampleYFilename}")
    else:
        np.savetxt(sampleYFilename, sampleY)


def readData(inputPath):
    """reads lowfi data"""
    sampleX = np.loadtxt(os.path.join(inputPath, "sampleX_bounds.txt"), encoding="utf8").T
    sampleY = AbstractDOE.ysFromFile(os.path.join(inputPath, "sampleY.pickle"))

    lowerBounds = np.min(sampleX, axis=1)
    upperBounds = np.max(sampleX, axis=1)

    return sampleX, sampleY, lowerBounds, upperBounds


def createSingleKriging(lowFiX, lowFiy, saveName, lb, ub, name, parameterNames, log10SampleY):
    kriging = Kriging(
        lowFiX,
        lowFiy,
        lowerBounds=lb,
        upperBounds=ub,
        parameterNames=parameterNames,
        resultNames=[name],
        doRegularizationParameterOpt=True,
        corr="cubic",
        regress="linear",
        log10SampleY=log10SampleY,
        # optThetaGlobalAttempts=1,
    )
    kriging.createSurrogateModel()
    kriging.save(saveName)
    with open(saveName.replace(".pickle", "_info.txt"), "w") as f:
        f.write(kriging.getKrigingDocumentation())

    log.info(f"kriging {name}: theta {kriging.theta},regularization {kriging.regularizationParameter}")

    return kriging


def getKrigings(samplesDir, surrogateDir, parameterNames, resultNamesIndexesLog10):
    sampleX, sampleYAll, lb, ub = readData(samplesDir)
    sampleYAll = np.array(sampleYAll)

    mms = KrigingsCaller()
    for resultName, index, useLog10 in resultNamesIndexesLog10:
        mmName = f"lh2_tank_{resultName}_v_1.1.pickle"
        readName = os.path.join(surrogateDir, f"tank_{resultName}.pickle")
        if os.path.exists(readName):
            mm = Kriging.load(readName)

            saveName = os.path.join(surrogateDir, mmName)
            mm.save(saveName)
            with open(saveName.replace(".pickle", "_info.txt"), "w") as f:
                f.write(mm.getKrigingDocumentation())
        else:
            writeData(surrogateDir, sampleX, sampleYAll)
            sampleY = np.array(sampleYAll[:, index], dtype=np.float64)
            if useLog10:
                sampleY = np.log10(sampleY)
            saveName = os.path.join(surrogateDir, mmName)
            mm = createSingleKriging(sampleX, sampleY, saveName, lb, ub, resultName, parameterNames, useLog10)
        mms.append(mm)

    # accuracy
    for mm in mms:
        crossSampleY = mm.getCrossValidationValues()
        sample1, sample2 = (
            (np.power(10, mm.sampleY), np.power(10, crossSampleY)) if mm.log10SampleY else (mm.sampleY, crossSampleY)
        )
        rmse = rootMeanSquareError(sample1, sample2)
        rmsre = relativeRootMeanSquareError(sample1, sample2)
        log.info(f"\nRMSE: {rmse}\nRMSRE: {rmsre}")

    return mms


def main():
    ap = ArgumentParser(
        description="Calculate tank properties based on geometry and pressure. It calculates frpMass[kg], volume[dm^3], area[m^2], length[mm]."
    )
    ap.print_usage = ap.print_help  # redirecting the print_usage method to the extended print_help method
    ap.add_argument("radius", metavar="r", type=float, help="radius of the tank [mm]")
    ap.add_argument("lZylByR", metavar="lZylByR", type=float, help="length of zylindrical part by radius [-]")
    ap.add_argument("dp", metavar="dp", type=float, help="delta pressure [MPa]")
    ap.add_argument(
        "-d",
        "--dir",
        dest="dir",
        help="Path to the input files (samples or created surrogates)",
        metavar="DIR",
    )
    options = ap.parse_args()
    shape = "isotensoid"
    # shape = 'circle'
    if options.dir:
        runDir = options.dir
        if not os.path.exists(runDir):
            raise FileNotFoundError(f'given directory "{runDir}" does not exist.')
    else:
        runDir = f"C:\\PycharmProjects\\delis\\tmp\\2020_10_doe_{shape}_41\\"
    r, lZylByR, dp = options.radius, options.lZylByR, options.dp

    mms = getKrigings(runDir, shape)

    frpMass, volume, area, length = mms([r, lZylByR, dp])
    print(frpMass, volume, area, length)


if __name__ == "__main__":
    main()



# Copyright (C) 2013 Deutsches Zentrum fuer Luft- und Raumfahrt(DLR, German Aerospace Center) <www.dlr.de>
"""
This class serves as exemplary controller class for surrogate calculations. Here all general
purpose input parameters are handled.
"""

import os

import matplotlib.pyplot as plt
import numpy as np
from patme.service.logger import log
from patme.service.stringutils import indent

from delismm.model.customsystemfunction import ExampleFunction
from delismm.model.doe import LatinHypercube, MonteCarlo
from delismm.model.metrics import maxError, rootMeanSquareError
from delismm.model.surrogate import Kriging


def runExample(doPlot=False):
    """doc"""
    log.info('Surrogate example starting in folder "{}"', os.getcwd())

    lowerBounds = {"testFunction": 0.0}
    upperBounds = {"testFunction": 1.0}

    log.info("Obtain data")
    # ===========================================================================
    # training data
    # ===========================================================================
    targetSystemFunction = ExampleFunction(lowerBounds, upperBounds)
    numberOfSamplePoints = 6
    hypercube = LatinHypercube(numberOfSamplePoints, len(lowerBounds))
    sampleX = targetSystemFunction.scaleToBounds(hypercube.sampleXNormalized)
    sampleY = hypercube.getY(targetSystemFunction)
    # ===========================================================================
    # test data
    # ===========================================================================
    testDoe = MonteCarlo(numberOfSamplePoints, len(lowerBounds))
    sampleXTest = targetSystemFunction.scaleToBounds(testDoe.sampleXNormalized)
    sampleYTest = hypercube.getY(targetSystemFunction)[:, 0]

    log.info("Create Kriging Model")
    kriging = Kriging(
        sampleX,
        sampleY,
        lowerBounds=targetSystemFunction.lowerBounds,
        upperBounds=targetSystemFunction.upperBounds,
        corr="gauss",
    )
    kriging.createSurrogateModel()
    log.info("kriging documentation:\n{}", kriging.getKrigingDocumentation())

    # =======================================================================
    # error metrics
    # =======================================================================
    errorFunctions = [rootMeanSquareError, maxError]
    # get kriging function values for comparison with sampleYTest
    sampleYTestMM = [kriging(sample)[0] for sample in sampleXTest.T]
    errors = [func(sampleYTest, sampleYTestMM) for func in errorFunctions]
    log.info("Error measures:\n{}", indent([[func.__name__, error] for func, error in zip(errorFunctions, errors)]))

    if doPlot and len(upperBounds) == 1:
        _, ax0 = plt.subplots(nrows=1)

        # plot target function
        sampleXTest = np.linspace(0.0, 1.0, 100)
        sampleYTest = [targetSystemFunction([poi]) for poi in sampleXTest]
        ax0.plot(sampleXTest, sampleYTest, "k:", label="Target function")

        # plot sample points
        ax0.plot(sampleX[0], sampleY, "ro", label="Sample Points")

        # plot kriging model
        sampleXKriging = np.linspace(0.0, 1.0, 100)
        sampleYKriging = [kriging(poi) for poi in sampleXKriging]
        ax0.plot(sampleXKriging, sampleYKriging, "--", label="Kriging Sample Points")

        ax0.set_title("Metamodel")
        ax0.set_xlabel(kriging.parameterNames[0] + " in []")
        ax0.set_ylabel("Test function value []")
        ax0.legend(
            loc="upper left",
        )
        plt.show()


if __name__ == "__main__":
    runExample(doPlot=True)
    # this will create a doe, sample values, a Metamodel and finally a diagram



from pydantic import BaseModel, PositiveInt


class SurrogateInfo(BaseModel):
    kriging_type: str
    model_name: str
    number_of_samples: int
    number_of_parameters: int
    condition: float
    correlation_function: str
    regression_function: str
    regularizationParameter: float
    optThetaGlobalLocalType: str
    optThetaGlobalAttempts: int
    optThetaMaxIter: int
    optThetaPrintout: bool
    correlationMatrixAlgo: str
    log10SampleX: list[bool]
    log10SampleY: bool
    parameterNames: list[str]
    lowerBounds: list[float]
    upperBounds: list[float]
    theta: list[float]



# -*- coding:utf-8 -*-
# Copyright (C) 2013 Deutsches Zentrum fuer Luft- und Raumfahrt(DLR, German Aerospace Center) <www.dlr.de>
"""
This module contains interfaces to arbitrary target functions.

Generally a target functions is defined by it's actual state which is given by the
actual system parameters. Additionally those parameters have lower and upper bounds
that are handled. Given samples may also be scaled to the bounds or to normalized
samples respectivly.
Lastly a target function must be defined for the system returning the system response
to the actual system state. The system response is returned as a list of arbitrary items.


.. note::
    The target function return value may be extended to return a list of
    system values and graphs in the future.


"""

import itertools
import os
import time
import traceback
from collections import OrderedDict

import numpy as np
from patme.service.logger import log
from patme.service.stringutils import indent

from delismm.service.exception import DelisMMError


class BoundsHandler:
    """This class handles the bounds of a target function. It is capable of converting them
    from the scaled to the non-scaled versions as well as writing and reading them."""

    def __init__(self, lowerBounds, upperBounds):
        """doc"""
        self.bounds = self.getBoundsMatrix(lowerBounds, upperBounds)
        self.lowerBounds = np.array(list(lowerBounds.values()))
        self.upperBounds = np.array(list(upperBounds.values()))

    @staticmethod
    def getBoundsMatrix(lowerBounds, upperBounds):
        """bounds is a list (p x 3), p: number of design variables
        each line is built like this: [ key, lower bound, upper bound]"""
        bounds = []
        for parameterName in lowerBounds.keys():
            if parameterName in upperBounds.keys():
                bounds.append([parameterName, lowerBounds[parameterName], upperBounds[parameterName]])
            else:
                raise DelisMMError("Key in lower bounds is not present in upper bounds. Key: " + str(parameterName))
        return bounds

    @staticmethod
    def getBoundsDicts(bounds):
        """bounds is a list (p x 3), p: number of design variables
        each line is built like this: [ key, lower bound, upper bound]
        returns the bounds as tuple with dictionaries: (lowerBounds, upperBounds)"""
        lowerBounds, upperBounds = OrderedDict([]), OrderedDict()
        for key, lowerBound, upperBound in bounds:
            lowerBounds[key] = lowerBound
            upperBounds[key] = upperBound
        return lowerBounds, upperBounds

    def boundsToFile(self, filename):
        """writes the given bounds and constants to a file. The file will be overwritten without warning!

        :param filename: name of the file and potentially the path where the bounds should be written
        """
        self.boundsToFileStatic(filename, self.bounds, self.customKeywordArgs)

    @staticmethod
    def boundsToFileStaticDict(filename, lowerBounds, upperBounds, constantsDict=None):
        """writes the given bounds and constants to a file. The file will be overwritten without warning!

        :param filename: name of the file and potentially the path where the bounds should be written
        """
        bounds = BoundsHandler.getBoundsMatrix(lowerBounds, upperBounds)
        BoundsHandler.boundsToFileStatic(filename, bounds, constantsDict)

    @staticmethod
    def boundsToFileStatic(filename, bounds, constantsDict):
        """writes the given bounds and constants to a file. The file will be overwritten without warning!

        :param filename: name of the file and potentially the path where the bounds should be written
        """
        if not os.path.exists(os.path.dirname(filename)) and os.path.dirname(filename) != "":
            raise DelisMMError("The directory for the given file does not exist: " + str(os.path.dirname(filename)))
        outString = "parameters:\n" + indent(bounds, delim="  ")
        if constantsDict:
            outString += "constants:\n" + indent(list(constantsDict.items()), delim="  ")
        with open(filename, "w") as f:
            f.write(outString)

    @staticmethod
    def boundsFromFileDict(filename):
        """see boundsFromFile

        :return: tuple of dictionaries: lowerBounds, upperBounds
        """
        bounds, constantValuesDict = BoundsHandler.boundsFromFile(filename)
        lowerBounds, upperBounds = OrderedDict(), OrderedDict()
        for key, lower, upper in bounds:
            lowerBounds[key] = lower
            upperBounds[key] = upper
        return lowerBounds, upperBounds, constantValuesDict

    @staticmethod
    def boundsFromFile(filename):
        """writes the given bounds to a file. The file will be overwritten without warning!

        :param filename: name of the file and potentially the path where the bounds should be written
        :return: bounds is a list (p x 3), p: number of design variables and , cosntantValuesDict as dictionary with constant values
        """
        if not os.path.exists(filename):
            raise DelisMMError("The file to read bounds does not exist: " + str(filename))
        with open(filename, "r") as f:
            lines = f.readlines()
        lineIter = iter(lines)
        bounds = []
        for line in lineIter:
            if line == "\n":
                continue
            if "parameters:" in line:
                continue
            if "constants:" in line:
                break
            key, lower, upper = line.split()
            bounds.append([key, float(lower), float(upper)])
        constantValuesDict = OrderedDict()
        for line in lineIter:
            key, value = line.split()
            if not value.isalpha():
                if "." in value:
                    value = float(value)
                else:
                    value = int(value)
            if "False" == value:
                value = False
            elif "True" == value:
                value = True
            constantValuesDict[key] = value
        return bounds, constantValuesDict

    def scaleToBounds(self, sampleXNormalized):
        """Method that scales given sample points from a normalized space to real bounds.

        :param sampleX: [n x p] array with samples
        """
        return BoundsHandler.scaleToBoundsStatic(sampleXNormalized, self.lowerBounds, self.upperBounds)

    @staticmethod
    def scaleToBoundsStaticDict(sampleXNormalized, lowerBounds, upperBounds):
        """Method that scales given sample points from a normalized space to real bounds.

        :param sampleX: [n x p] array with samples
        :param lowerBounds: dict with lower bounds of each parameter
        :param upperBounds: dict with upper bounds of each parameter
        """
        return BoundsHandler.scaleToBoundsStatic(
            sampleXNormalized, np.array(list(lowerBounds.values())), np.array(list(upperBounds.values()))
        )

    @staticmethod
    def scaleToBoundsStatic(sampleXNormalized, lowerBoundsArray, upperBoundsArray):
        """Method that scales given sample points from a normalized space to real bounds.

        :param sampleX: [n x p] array with samples
        """
        lowerBoundsArray = np.array(lowerBoundsArray)
        upperBoundsArray = np.array(upperBoundsArray)
        sampleXNormalized = np.array(sampleXNormalized)
        return (lowerBoundsArray + sampleXNormalized.T * (upperBoundsArray - lowerBoundsArray)).T

    def scaleTo01(self, sampleX):
        """Method that scales given sample points to given bounds.

        :param sampleX: [p x n] array with samples
        """
        return ((sampleX.T - self.lowerBounds) / (self.upperBounds - self.lowerBounds)).T

    @staticmethod
    def scaleTo01Static(sampleX, lowerBoundsArray, upperBoundsArray):
        """Method that scales given sample points to given bounds.
        :param lowerBoundsArray: array with lower bounds of each parameter
        :param upperBoundsArray: array with upper bounds of each parameter
        :param sampleX: [n x p] array with samples
        """
        return ((sampleX.T - lowerBoundsArray) / (upperBoundsArray - lowerBoundsArray)).T

    def _getParameterNames(self):
        """doc"""
        return [bound[0] for bound in self.bounds]

    parameterNames = property(fget=_getParameterNames)


class AbstractTargetFunction(BoundsHandler):
    """classdocs"""

    name = "AbstractTargetFunction"

    def __init__(self, lowerBounds, upperBounds, resultNames=None):
        """doc"""
        BoundsHandler.__init__(self, lowerBounds, upperBounds)
        self.p = len(self.bounds)
        self.customKeywordArgs = {}
        """This is a dictionary that may contain custom keyword arguments for the model run.
        With this, one can overwrite the defaults, eg. for panels which are defined in self._getPanel"""

        if not resultNames:
            resultNames = ["resultName"]
        if isinstance(resultNames, str) or not hasattr(resultNames, "__iter__"):
            resultNames = [resultNames]
        self.resultNames = resultNames
        """string or list with the names of the keys returned by self.__call__"""

        self.runDir = None
        self.storeResults = False
        """Flag if the calculation designs and their results should be stored in a file."""
        self.doParallelization = []
        """Parallelization for various purposes can be any combination of "clusterqueue, local, async" or nothing.

        1. "clusterqueue" splits the job according to the actual number of available cluster jobs -1
        2. "local" allows to perform local parallelization according to 3//4 of the number of available cores.
           Local may be combined with clusterqueue, whereas the latter is used first, parallelizing locally to generate
           cluster jobs and performs a second parallelization on the cluster
        3. "async" performs the parallelization locally but uses a defined function to get the maximal number of
           allowed executions. This function is called "getNumberOfNewJobs" and must be
           implemented in the inheriting targetFunc.
           "async" may not be used with "local" at the same time but with clusterqueue.
        """
        self.asyncWaitTime = 0
        """Time in seconds to wait after starting a new async job"""
        self.asyncMaxProcesses = None
        """Maximum number of parallel processes for asynchronous calculations. Must be set in subclasses. """
        self.doRemoteCall = False
        """Flag if the samples should be calculated on the fa-institute cluster"""
        self.allowFailedSample = False
        """In case of an exception raised in the calculation of one sample,
        - the process stops if False
        - continues calculating the next sample if True

        Especially async parallel processes must not fail with an exception since they can create a deadlock.
        If you call samplecalculator.getY() in async mode, this flag should be set to true"""

    def __call__(self, parameters):
        """
        Function that is called to retrieve the actual system response.
        :param parameters: iterable containing the actual system parameters in the same order as self.bounds
        :return: list with system responses
        """
        try:
            result = self._call(np.array(parameters))
        except Exception as e:
            if self.allowFailedSample:
                log.error("\n".join(["Exception raised with this message:", str(e), traceback.format_exc()]))
                result = None
            else:
                raise
        if not isinstance(result, str) and not hasattr(result, "__iter__"):
            # create a list from the result
            result = [result]
        return result

    def _call(self, parameters):
        """see __call__ description

        :return: float or list as system response. The floats will be put in a list in self.__call__!
        """
        raise NotImplementedError("This method must be implemented in a subclass")

    def getNumberOfNewJobs(self):
        """This method is used for asynchronous parallel executions. It returns the number of parallel
        jobs which may vary over time."""
        raise NotImplementedError("This method must be implemented in a subclass")

    def _writeResultToFile(self, result, modelInfoString):
        """method that writes results and the model information to a file.
        Usually the file is kept open during creation of sample points."""
        filename = self.name + "_designs_results.txt"
        if self.runDir is not None:
            filename = os.path.join(self.runDir, filename)
        with open(filename, "a") as f:
            f.write(modelInfoString + "\n")
            f.write("Result: " + str(result) + "\n\n")


class ExampleFunction(AbstractTargetFunction):
    """function from Forrester: engineering design via surrogate modelling (ch. 8.2)"""

    def __init__(self, lowerBounds, upperBounds):
        AbstractTargetFunction.__init__(self, lowerBounds, upperBounds)
        self.name = self.__class__.__name__

    def _call(self, parameters):
        parameters = np.array(parameters)
        return np.sum((6.0 * parameters - 2) ** 2 * np.sin(12 * parameters - 4))


class ExampleFunctionModified(ExampleFunction):
    """See report of Sauerbrei p.38
    function from Forrester: engineering design via surrogate modelling (ch. 8.2)"""

    def __init__(self, lowerBounds, upperBounds):
        ExampleFunction.__init__(self, lowerBounds, upperBounds)

    def _call(self, parameters):
        parameters = np.array(parameters)
        return np.sum(0.5 * ExampleFunction._call(self, parameters) + 10 * (parameters - 0.5) - 5)


class ExampleFunctionAsyncParallelization(ExampleFunction):
    """This is an examplary function for async parallelization"""

    def __init__(self, lowerBounds, upperBounds):
        super().__init__(lowerBounds, upperBounds)
        self.doParallelization = ["async"]
        self.asyncWaitTime = 0.001
        self.asyncMaxProcesses = 4

        self._newJobsScheme = [1, 0]
        """only for testing"""

    def _setNumberOfNewJobsGenerator(self):
        """doc"""
        generator1 = range(3, 0, -1)
        generator2 = itertools.cycle(self._newJobsScheme)
        generator = itertools.chain(generator1, generator2)
        for value in generator:
            yield value

    def getNumberOfNewJobs(self):
        """Creates a generator that returns [3,2,1,1,0,1,0...] as example"""
        if not hasattr(self, "numberOfNewJobsGenerator"):
            self.numberOfNewJobsGenerator = self._setNumberOfNewJobsGenerator()
        return next(self.numberOfNewJobsGenerator)

    def _call(self, parameters):
        """delays the execution randomly in order to test correct order of the samples"""
        time.sleep(np.random.rand() * 0.1)
        return super()._call(parameters)


class ExampleLinearCoupledFunction(AbstractTargetFunction):
    """
    calculates: sum(x_i*i) + sum(x_j*j*x_(j+1)*(j+1))
    where i = 1, .. ,len(self.paramters)
    and   j = 1, .. ,len(self.paramters)-1
    """

    def __init__(self, lowerBounds, upperBounds):
        AbstractTargetFunction.__init__(self, lowerBounds, upperBounds)
        self.name = "ExampleLinearCoupledFunction"

    def _call(self, parameters):
        scaledParams = parameters * np.arange(1, len(parameters) + 1)
        couplingParams = list(zip(scaledParams[:-1], scaledParams[1:]))
        return np.sum(scaledParams) + np.sum(couplingParams)


class ExampleQuadraticCoupledFunction(AbstractTargetFunction):
    """
    calculates: sum(x_i^2*i) + sum(x_j*x_(j+1))
    where i = 1, .. ,len(self.paramters)
    and   j = 1, .. ,len(self.paramters)-1
    """

    def __init__(self, lowerBounds, upperBounds):
        AbstractTargetFunction.__init__(self, lowerBounds, upperBounds)
        self.name = "ExampleQuadraticCoupledFunction"

    def _call(self, parameters):
        scaledParams = parameters * parameters * np.arange(1, len(parameters) + 1)
        couplingParams = list(zip(parameters[:-1], parameters[1:]))
        return np.sum(scaledParams) + np.sum(couplingParams)


class BraninFunction(AbstractTargetFunction):
    """branin function equation 12 of  D. J. J. Toal, "Some considerations regarding the use of multi-fidelity Kriging in the construction of surrogate models,"  2015"""

    def __init__(self, lowerBounds, upperBounds):
        AbstractTargetFunction.__init__(self, lowerBounds, upperBounds)
        self.name = self.__class__.__name__

    def _call(self, parameters):
        x1, x2 = parameters
        return (
            (x2 - 5.1 * x1**2 / 4 / np.pi**2 + 5.0 * x1 / np.pi - 6) ** 2 + 10 * (1 - 1.0 / 8 / np.pi) * np.cos(x1) + 10
        )


class BraninFunctionModified(BraninFunction):
    """branin function modified equation 13 of  D. J. J. Toal, "Some considerations regarding the use of multi-fidelity Kriging in the construction of surrogate models,"  2015"""

    def __init__(self, lowerBounds, upperBounds):
        AbstractTargetFunction.__init__(self, lowerBounds, upperBounds)
        self.name = self.__class__.__name__
        self.a1 = 0.0  # [0,1] is used in the paper

    def _call(self, parameters):
        fBranin = super()._call(parameters)
        x1, x2 = parameters
        return fBranin - (self.a1 + 0.5) * (x2 - 5.1 * x1**2 / 4 / np.pi**2 + 5.0 * x1 / np.pi - 6) ** 2


class JoinedTargetFunction(AbstractTargetFunction):
    """Class for arbitrary target functions. They can be given as function parameter to the constructor"""

    def __init__(self, targetFunction, lowerBounds, upperBounds, name="Joined Target Function"):
        AbstractTargetFunction.__init__(self, lowerBounds, upperBounds)
        self.name = name
        self.targetFunction = targetFunction

    def __call__(self, pointOfInterest):
        """See AbstractKriging for documentation

        returns the value of self.targetFunction()"""
        return self.targetFunction(pointOfInterest)


class RegressionCorrelationCommons:
    """classdoc"""

    def __init__(self, kriging):
        """doc"""
        self.kriging = kriging
        self.name = self.kriging.name + self._name

    def __call__(self, pointOfInterest):
        """See AbstractKriging for documentation

        returns the stochastic correlation part of the kriging model"""
        self.kriging.checkModelCreation()
        return [self._call(pointOfInterest)]

    def _getLowerBounds(self):
        """doc"""
        return self.kriging.lowerBounds

    def _getUpperBounds(self):
        """doc"""
        return self.kriging.upperBounds

    def _getParameterNames(self):
        """doc"""
        return self.kriging.parameterNames

    parameterNames = property(fget=_getParameterNames)
    lowerBounds = property(fget=_getLowerBounds)
    upperBounds = property(fget=_getUpperBounds)


class KrigingMse(RegressionCorrelationCommons):
    """Target function for kriging mse. It is used as kriging(x)+-mse(x). The +- is identified by the flag isUpper"""

    _name = "_regr"

    def __init__(self, kriging, isUpper=True):
        """see base class description

        :param isUpper: flag if the upper or the lower MSE bound should be created
        """
        self._name = "_mse_up" if isUpper else "_mse_low"
        self._isUpper = 1 if isUpper else -1
        RegressionCorrelationCommons.__init__(self, kriging)

    def _call(self, pointOfInterest):
        """See AbstractKriging for documentation

        returns the mse of the kriging model"""
        pointOfInterestNormalized = BoundsHandler.scaleTo01Static(
            np.array(pointOfInterest), self.lowerBounds, self.upperBounds
        )
        mse = self._isUpper * self.kriging.getMSEofPredictor(pointOfInterestNormalized)
        return self.kriging(pointOfInterestNormalized)[0] + mse


class KrigingRegression(RegressionCorrelationCommons):
    """Target function for the regression function of a given kriging model"""

    _name = "_regr"

    def _call(self, pointOfInterest):
        """See AbstractKriging for documentation

        returns the regression part of the kriging model"""
        pointOfInterestNormalized = BoundsHandler.scaleTo01Static(
            np.array(pointOfInterest), self.lowerBounds, self.upperBounds
        )
        return self.kriging.regressionFunction(pointOfInterestNormalized)


class KrigingCorrelation(RegressionCorrelationCommons):
    """Target function for the regression function of a given kriging model"""

    _name = "_corr"

    def _call(self, pointOfInterest):
        """See AbstractKriging for documentation

        returns the stochastic correlation part of the kriging model"""
        pointOfInterestNormalized = BoundsHandler.scaleTo01Static(
            np.array(pointOfInterest), self.lowerBounds, self.upperBounds
        )
        corrVec = self.kriging.getCorrelationVector(pointOfInterestNormalized)
        result = corrVec.T @ self.kriging.v
        return result[0]


class KrigingParallelization:
    """classdoc"""

    def __init__(self, kriging, args, doRemoteCall, runDir=None):
        """doc"""
        self.kriging = kriging
        """the kriging model to be used"""
        self.args = args
        """arguments to the theta optimization function"""
        self.runDir = runDir
        self.doRemoteCall = doRemoteCall
        self.doParallelization = ["clusterqueue"]


class ThetaOptimizerKriging(KrigingParallelization):
    """Target function for the regression function of a given kriging model"""

    name = "krigOpt"

    def __call__(self, pointOfInterest):
        """See AbstractKriging for documentation

        returns the stochastic correlation part of the kriging model"""
        if pointOfInterest != [0]:
            raise DelisMMError("This method should not get any used parameter input")
        return self.kriging._getGlobalOptimum(*self.args)


class CrossValidationValuesCalculatorKriging(KrigingParallelization):
    """Target function for the regression function of a given kriging model"""

    name = "krigCV"

    def __call__(self, pointOfInterest):
        """returns the crossvalidation value at a certain point"""
        return self.kriging.crossValidation(pointOfInterest, *self.args)


class AbsError:
    """Target function for the error function of a given model

    calculates: model2 - model1"""

    def __init__(self, model1, model2):
        """
        :param model1: first model
        :param targetFunction: target function that was used to create the kriging model
        """
        self.model1 = model1
        self.model2 = model2

    def __call__(self, pointOfInterest):
        """See AbstractKriging for documentation"""
        return self.model2(pointOfInterest) - self.model1(pointOfInterest)

    def _getLowerBounds(self):
        """doc"""
        return self.model1.lowerBounds

    def _getUpperBounds(self):
        """doc"""
        return self.model1.upperBounds

    lowerBounds = property(fget=_getLowerBounds)
    upperBounds = property(fget=_getUpperBounds)


class RelError(AbsError):
    """Target function for the error function of a given kriging model

    calculates: 1 - (min(model1,model2) / max(model1,model2))"""

    def __call__(self, pointOfInterest):
        """See AbstractKriging for documentation"""
        s = np.abs(self.model2(pointOfInterest))
        k = np.abs(self.model1(pointOfInterest))
        return 1 - np.min([s, k]) / np.max([s, k])


class ModelDimensionalityTransformer:
    """This class is intended to encapsulate targetfunctions and metamodels in order
    to reduce or increase the number of parameters for a transformation in dimensionality using constants.

    Parameters that are unused(self has more params than model), will be simply omitted.
    Parameters that are added(self has less params than model) to the model, will be introduced as constants.
    Also a mixture of reducing and adding parameters is feasible.

    See test case for examples.
    """

    def __init__(self, model, modelKeys, newModelKeys, constantsValueDict=None):
        """Transforms parameters from newModekKeys --> modelKeys"""
        self.model = model
        differenceKeys = set(modelKeys).difference(newModelKeys)
        if differenceKeys and constantsValueDict is None:
            raise DelisMMError(
                "constantsValueDict must be given as dict since at least one constant values for these keys are reqired: "
                + str(differenceKeys)
            )

        self.parameterTrafo = np.zeros((len(modelKeys), len(newModelKeys)))
        self.parameterConstants = np.zeros((len(modelKeys),))
        for modelKeyIndex, modelKey in enumerate(modelKeys):
            if modelKey in newModelKeys:
                self.parameterTrafo[modelKeyIndex, newModelKeys.index(modelKey)] = 1
            else:
                self.parameterConstants[modelKeyIndex] = constantsValueDict[modelKey]
        print()
        print(modelKeys)
        print(self.parameterTrafo)
        print(self.parameterConstants)

    def __call__(self, parameters):
        """
        Function that is called to retrieve the actual system response.
        :param parameters: iterable containing the actual system parameters in the same order as self.bounds
        :return: float, system response
        """
        parameters = np.array(parameters)
        return [self.model(np.dot(self.parameterTrafo, parameters) + self.parameterConstants)]


if __name__ == "__main__":
    if 0:
        fe = BraninFunction(
            lowerBounds=OrderedDict([("x1", -5.0), ("x2", 0.0)]), upperBounds=OrderedDict([("x1", 10.0), ("x2", 15.0)])
        )
        fc = BraninFunctionModified(
            lowerBounds=OrderedDict([("x1", -5.0), ("x2", 0.0)]), upperBounds=OrderedDict([("x1", 10.0), ("x2", 15.0)])
        )
        print(fe((-np.pi, 12.275)))
        print(fe((np.pi, 2.275)))
        print(fe((9.42478, 2.475)))
        from delismm.model.doe import FullFactorialDesign
        from delismm.model.metrics import correlation, rootMeanSquareError
        from delismm.model.samplecalculator import getY

        doe = FullFactorialDesign(20**2, 2)
        sampleX = fe.scaleToBounds(doe.sampleXNormalized)
        rmses, r2s = [], []
        a1s = np.linspace(0.0, 1.0, 51)
        for a1 in a1s:
            fc.a1 = a1

            sye = getY(sampleX, fe)
            syc = getY(sampleX, fc)
            rmses.append(rootMeanSquareError(sye, syc))
            r2s.append(correlation(sye, syc) ** 2)

        print(indent(zip(rmses, r2s)))

        import matplotlib.pyplot as plt

        fig, ax1 = plt.subplots()
        ax1.plot(a1s, r2s, color="r", label="r2")
        ax1.set_xlabel("A1")
        # Make the y-axis label, ticks and tick labels match the line color.
        ax1.set_ylabel("r2")
        plt.legend(loc="upper left")

        ax2 = ax1.twinx()
        ax2.plot(a1s, rmses, color="b", label="rmse")
        ax2.set_ylabel("RMSE")
        plt.legend()
        plt.show()



"""client for testing"""

import requests

if __name__ == "__main__":
    server = "https://delismm.nimbus.dlr.de/"
    surrogateName = "tank_isotensoid_frpMass[kg]_v_1.0"

    print("############ Surrogate list #############")
    endpoint = "list_surrogates"
    request = f"{server}/{endpoint}"
    r = requests.get(request)
    print(request, r.status_code, r.text)

    print("############ Surrogate info #############")
    endpoint = f"surrogate_info/{surrogateName}"
    request = f"{server}/{endpoint}"
    r = requests.get(request)
    print(request, r.status_code, r.text)

    print("############ Surrogate call #############")
    parameters = ["2000", "1000", "0.5"]  # dcyl, lcyl, pressure
    request = (
        f"https://delismm.nimbus.dlr.de/call/?surrogate_name={surrogateName}&"
        f"parameters={parameters[0]}&parameters={parameters[1]}&parameters={parameters[2]}"
    )
    r = requests.get(request)
    print(request, r.status_code, r.text)


if __name__ == "__main__":
    from delismm.api import getSurrogate

    surrogateName = "tank_isotensoid_frpMass[kg]_v_1.0"
    s = getSurrogate(surrogateName)
    p = [2000, 1000, 0.5]
    print(s(p))



""""""

import numpy as np
from patme.service.systemutils import getTimeString
from scipy import interpolate
from scipy.interpolate import RegularGridInterpolator

from delismm.model.customsystemfunction import BoundsHandler, KrigingCorrelation, KrigingRegression
from delismm.model.surrogate import HierarchicalKriging, Kriging


class RegularGridInterpolatorContainer:
    """RegularGridInterpolator should behave like a kriging model"""

    def __init__(self, rgi, lb, ub):
        self.rgi = rgi
        self.lb = lb
        self.ub = ub

    def __call__(self, xi, method="linear"):
        return self.rgi([xi])[0]

    def callNormalized(self, poiNomalized):
        return [self(BoundsHandler.scaleToBoundsStatic(poiNomalized, self.lb, self.ub))]


def plotMetamodel(plotter, mm, lb, ub):
    plotSampeles = 150
    x = np.linspace(lb[0], ub[0], plotSampeles)
    y = np.linspace(lb[1], ub[1], plotSampeles)
    xv, yv = np.meshgrid(x, y)
    xv, yv = xv.flatten(), yv.flatten()
    z = np.array([mm([xVal, vVal]) for xVal, vVal in zip(xv, yv)])
    plotter.plotContour(xv, yv, z)


def plotHighFiSamples(plotter, vmm, mm):
    hfx = vmm.sampleX.T
    hfy = vmm.sampleY

    diffMmVmm = hfy - np.array([mm(poi) for poi in hfx])

    plotter.plotScatter(hfx[:, 0], hfx[:, 1], pointColors=diffMmVmm, label="hifi-lowfi")


def getInterp(inputPath, plotter):
    """returns interpolation function for lowfi data"""
    lowFi = np.loadtxt(inputPath + "082 C2 260 158_AOIcartCorr.txt")
    samplesPerDirection = int(np.sqrt(lowFi.shape[0]))

    plotter.plotContour(lowFi[:, 0], lowFi[:, 1], lowFi[:, 2])

    if 0:
        f = interpolate.interp2d(lowFi[:, 0], lowFi[:, 1], lowFi[:, 2], kind="cubic")
    else:
        data = lowFi[:, 2].reshape((samplesPerDirection, samplesPerDirection)).T
        f = RegularGridInterpolator(
            (lowFi[:samplesPerDirection, 0], lowFi[::samplesPerDirection, 1]),
            data,
            bounds_error=False,
            fill_value=0.0,
        )

    return f, np.min(lowFi[:, :2], axis=0), np.max(lowFi[:, :2], axis=0)


def readData(inputPath, plotter, useEvery):
    """reads lowfi data"""
    lowFi = np.loadtxt(inputPath + "082 C2 260 158_AOIcartCorr.txt")
    samplesPerDirection = int(np.sqrt(lowFi.shape[0]))

    plotter.plotContour(lowFi[:, 0], lowFi[:, 1], lowFi[:, 2])

    lowFiReshaped = np.reshape(
        lowFi, (int(lowFi.shape[0] / samplesPerDirection), int(lowFi.shape[0] / samplesPerDirection), 3)
    )
    lowFiReduced = lowFiReshaped[::useEvery, ::useEvery, :]
    lowFiX = np.reshape(lowFiReduced, (int((samplesPerDirection / useEvery) ** 2), 3))[:, :2].T
    lowFiy = np.reshape(lowFiReduced, (int((samplesPerDirection / useEvery) ** 2), 3))[:, 2]

    lowerBounds = np.min(lowFiX, axis=1)
    upperBounds = np.max(lowFiX, axis=1)

    return lowFiX, lowFiy, lowerBounds, upperBounds


def readDataHiFi(inputPath, plotter):
    """read highfi data"""
    highFi = np.loadtxt(inputPath + "082_Thermoelemente_Auswahl.txt")

    plotter.plotScatter(highFi[:, 0], highFi[:, 1])

    return highFi[:, :2], highFi[:, 3]


def createKrigingLowFi(lowFiX, lowFiy, savePath, lb, ub):
    kriging = Kriging(
        lowFiX,
        lowFiy,
        lowerBounds=lb,
        upperBounds=ub,
        parameterNames=["x", "y"],
        doRegularizationParameterOpt=False,
        optThetaGlobalAttempts=2,
    )
    kriging.createSurrogateModel()  # theta=[0.55175332, 0.73914592])
    kriging.save(f"kriging_model_{kriging.n}.pickle", savePath)

    print(kriging.theta, kriging.regularizationParameter)

    return kriging


def main():
    from delis.service.dataplotter import PlotGraphMatplotlib

    plotter = PlotGraphMatplotlib(shapeSubplots=(2, 3))
    inputPath = "C:\\PycharmProjects\\delis\\tmp\\temperatures\\"
    useEvery = 15

    if 0:
        lowFiX, lowFiy, lb, ub = readData(inputPath, plotter, useEvery)
        mm = createKrigingLowFi(lowFiX, lowFiy, inputPath, lb, ub)
        lb, ub = mm.lowerBounds, mm.upperBounds
    else:
        rgi, lb, ub = getInterp(inputPath, plotter)
        mm = RegularGridInterpolatorContainer(rgi, lb, ub)

    # highFi
    hfx, hfy = readDataHiFi(inputPath, plotter)
    plotter.activeNextSubplot()

    plotMetamodel(plotter, mm, lb, ub)
    plotter.activeNextSubplot()

    # create hf mm
    vmm = HierarchicalKriging(
        hfx.T,
        hfy,
        mm,
        lowerBounds=lb,
        upperBounds=ub,
        parameterNames=["x", "y"],
        doRegularizationParameterOpt=False,
        optThetaGlobalAttempts=2,
    )
    vmm.createSurrogateModel(theta=[3, 3])
    mmAtHf = [mm(poi) for poi in hfx]
    vmmAtHf = [vmm(poi) for poi in hfx]
    vmmMinMm = [vmm(poi) - mm(poi) for poi in hfx]

    plotMetamodel(plotter, vmm, lb, ub)
    plotHighFiSamples(plotter, vmm, mm)
    plotter.activeNextSubplot()

    # plot correlation model
    krigingCorr = KrigingCorrelation(vmm)
    plotMetamodel(plotter, krigingCorr, lb, ub)
    plotHighFiSamples(plotter, vmm, mm)
    plotter.activeNextSubplot()

    # plot regression model
    krigingReg = KrigingRegression(vmm)
    plotMetamodel(plotter, krigingReg, lb, ub)
    plotHighFiSamples(plotter, vmm, mm)
    plotter.activeNextSubplot()

    plotter.legend()
    plotter.show()


if __name__ == "__main__":
    main()
